==> dinfo-1.0.tar.gz: dinfo-1.0/dinfo.pl <==
#!/usr/bin/perl -w
# Look up info for a domain
# $Fuller: myprograms/dinfo/dinfo.pl,v 1.6 2003/06/07 23:21:59 fullermd Exp $

use Net::DNS;
use strict;

# We'll use these a lot
my $i;
my %srvcache;
my %namecache;
my $ind = "  ";

# Create a resolver connection
my $res = Net::DNS::Resolver->new;
# Don't recurse
$res->recurse(0);

# First, some variables
# root server hints
my @roothints = (
		"128.9.0.107",
		"192.58.128.30",
		"193.0.14.129",
		"198.32.64.12",
		"202.12.27.33",
		"192.36.148.17",
		"192.203.230.10",
		"128.8.10.90",
		"198.41.0.4",
		"128.63.2.53",
		"192.33.4.12",
		"192.112.36.4",
		"192.5.5.241" );

# Look up the current root servers
$res->nameservers(randomize_list(@roothints));
my $query = $res->query(".", "NS");
if($query)
{
	foreach my $rr ($query->additional)
	{
		next unless (($rr->type eq "A")||($rr->type eq "AAAA"));
		addcachedserver(".", $rr->name, $rr->address);
		cachename($rr->name, $rr->type, $rr->address);
	}
} else {
	print "Error: Looking up root servers failed\n";
	exit(1);
}

# Loop over the domain we've requested
my $domain;
my @domain_servers;
$i=0;
while($domain = $ARGV[$i])
{
	# Get the info
	print "- Researching $domain...\n";

	# Full domain
	my %domitself = getdinfo("$domain");
	if(!%domitself)
	{ next; }

	# If it's a CNAME, chase it
	my @domcnameaddr;
	if(defined($domitself{"CNAME"}))
	{
		my $cnameto = $domitself{"CNAME"};
		@domcnameaddr = getinfo($cnameto, "A");
		if(!@domcnameaddr)
		{ @domcnameaddr = getinfo($cnameto, "AAAA"); }
	}

	# Look up its mailservers if we haven't already
	my %mxsrvs;
	if(defined($domitself{"MX"}))
	{
		foreach my $pref (keys %{$domitself{"MX"}})
		{
			map
			{
				my $srvname = $_;
				my @addrs;
				my $cnret = getinfo($srvname, "CNAME");
				if($cnret)
				{
					# It's a CNAME.  Chase it.
					push(@addrs, getinfo($cnret, "A"));
					push(@addrs, getinfo($cnret, "AAAA"));
					if(length(@addrs)>0)
					{
						$cnret .= " => " . addrs2string(@addrs);
					}
					$mxsrvs{$pref}->{$srvname} = $cnret;
				}
				else
				{
					push(@addrs, getinfo($srvname, "A"));
					push(@addrs, getinfo($srvname, "AAAA"));
					$mxsrvs{$pref}->{$srvname} = addrs2string(@addrs);
				}
			} @{$domitself{"MX"}->{$pref}};
		}
	}

	# www.$domain
	my @wwwinfo;
	my $wwwcname="";
	my @wwwcnameaddr;
	# Check for CNAME first
	$wwwcname = getinfo("www.$domain", "CNAME");
	if($wwwcname)
	{
		@wwwcnameaddr=getinfo($wwwcname, "A");
		if(!@wwwcnameaddr)
		{ @wwwcnameaddr=getinfo($wwwcname, "AAAA"); }
		if(!@wwwcnameaddr)
		{ @wwwcnameaddr = (); }
		@wwwinfo = ('foo');
	}
	else
	{
		push(@wwwinfo, getinfo("www.$domain", "A"));
		push(@wwwinfo, getinfo("www.$domain", "AAAA"));
	}

	# Show the info
	# Host A/AAAA/CNAME
	if( (defined($domitself{"A"})) || (defined($domitself{"AAAA"})) )
	{
		my $srvaddrs="";
		$srvaddrs .= addrs2string(@{$domitself{"A"}});
		$srvaddrs .= addrs2string(@{$domitself{"AAAA"}});
		print "${ind}o Host $domain has address ($srvaddrs)\n";
	}
	elsif(defined($domitself{"CNAME"}))
	{
		# CNAME'd instead
		my $cnameto;
		my $domcnameaddr="";
		$domcnameaddr .= addrs2string(@domcnameaddr);
		$cnameto = $domitself{"CNAME"};
		print "${ind}o Host $domain is a CNAME to $cnameto ($domcnameaddr)\n";
	}

	# Webserver
	if(@wwwinfo)
	{
		if($wwwcname)
		{
			my $wwwcnameaddr="";
			$wwwcnameaddr .= addrs2string(@wwwcnameaddr);
			print "${ind}o Host www.$domain is a CNAME to $wwwcname "
					. "($wwwcnameaddr)\n";
		}
		else
		{
			my $srvaddrs="";
			$srvaddrs .= addrs2string(@wwwinfo);
			print "${ind}o Host www.$domain has address ($srvaddrs)\n";
		}
	}

	# MX servers
	if(%mxsrvs)
	{
		print "${ind}o Mail exchangers:\n";
		foreach my $pref (sort { $a <=> $b } keys %mxsrvs)
		{
			my $firstone=1;
			foreach my $host (sort keys %{$mxsrvs{$pref}})
			{
				if($firstone)
				{ printf "${ind}${ind}* Pri=%3d ", $pref; $firstone=0; }
				else
				{ print  "${ind}${ind}          "; }
				printf "$host (%s)\n", $mxsrvs{$pref}->{$host};
			}
		}
	}

	# Nameservers (delegated)
	print "${ind}o Nameservers (from lower level nameservers):\n";
	foreach my $hname (sort keys %{$srvcache{"$domain."}})
	{
		my $nsaddrs="";
		$nsaddrs .= addrs2string(@{$srvcache{"$domain."}->{$hname}});
		print "${ind}${ind}* $hname ($nsaddrs)\n";
	}

	# Nameservers (in zomefile)
	print "${ind}o Nameservers (in server zonefile):\n";
	map
	{
		my $nsaddrs="";
		$nsaddrs .= addrs2string(@{$domitself{'additional'}->{$_}->{"A"}});
		$nsaddrs .= addrs2string(@{$domitself{'additional'}->{$_}->{"AAAA"}});
		print "${ind}${ind}* $_ ($nsaddrs)\n";
	} sort @{$domitself{'NS'}};

	# SOA
	if($domitself{'contact'})
	{
		my $caddr = $domitself{'contact'};
		$caddr =~ s/\./@/;
		print "${ind}o SOA contact address: $caddr\n";
	}


	# Move on down the road.
	$i++;
}


# FUNCTIONS
# Randomize a list
sub randomize_list
{
	my(@thislist) = @_;
	my $nitems = @thislist;
	my $i;
	my $randret;
	my @toret;

	for($i=0 ; $i < $nitems ; $i++)
	{
		while($randret=rand($nitems))
		{
			if( (length($thislist[$randret]))>0 )
			{
				$toret[$i] = $thislist[$randret];
				$thislist[$randret]="";
				last;
			}
		}
	}

	return(@toret);
}


# Get all available info for the given domain.  We cache nameservers
# along the way, which is useful if we're looking up multiple domains.
# Be nice to the networks, people!
sub getdinfo
{
	my($dname) = @_;
	my $nametrim;
	my %toret;
	my $haveanswer=0;
	my $dummy;

	# Start moving up
	while($haveanswer==0)
	{
		# Randomize to try and spread load
		$res->nameservers(randomize_list(getcachedservers($dname)));

		my $query = $res->send("$dname.", "ANY");
		if(!$query)
		{
			print "Query failed.  Status:\n";
			$res->print;
			print "\n";
			return;
		}

		# First see what response we have
		my $header = $query->header;
		my $response = $header->rcode;
		if($response ne "NOERROR")
		{
			# Some kinda error.  Just bail.
			my $errhost = $query->answerfrom();
			print "Error looking up domain info for '$dname' "
					. "from $errhost: $response\n";
			return;
		}

		# Check what we have
		# If this is authoritative, we're done.  Else, we're ready to
		# loop back around.
		if($header->aa==1)
		{
			# Yay, authoritative.
			# We're getting our answers.  Cool beans.
			# in the chain.  Cool beans.

			foreach my $rr ($query->answer)
			{
				my $type = $rr->type;

				if( ($type eq "A") || ($type eq "AAAA") )
				{
					push(@{$toret{$type}}, $rr->address);
					cachename($rr->name, $rr->type, $rr->address);
				}
				elsif($type eq "NS")
				{
					push(@{$toret{$type}}, $rr->nsdname);
					cachename($rr->name, $rr->type, $rr->nsdname);
				}
				elsif($type eq "CNAME")
				{
					$toret{$type} = $rr->cname;
					cachename($rr->name, $rr->type, $rr->cname);
				}
				# MX and SOA records don't get cached because I can't be
				# bothered to generalize enough.
				elsif($type eq "SOA")
				{ $toret{'contact'} = $rr->rname; }
				elsif($type eq "MX")
				{ push(@{$toret{$type}->{$rr->preference}}, $rr->exchange); }
				# Forget the other types for now
			}

			# Get info from our "additional" section.
			foreach my $rr ($query->additional)
			{
				my $type = $rr->type;
				my $rname = $rr->name;

				if( ($type eq "A") || ($type eq "AAAA") )
				{
					push(@{$toret{'additional'}->{$rname}->{$type}},
							$rr->address);
				}
				# Forget the other types for now
			}

			# That should be about all we care about
			$haveanswer=1;
		}
		else
		{
			# Non-authoritative (but still non-error) answer.  This
			# should be a referral to an upstream DNS server.  So, just
			# check the 'authority' section.
			my $domname;
			my @srvrs;

			foreach my $rr ($query->authority)
			{
				next unless $rr->type eq "NS";
				$domname = $rr->name;
				push(@srvrs, $rr->nsdname);
			}

			if($header->arcount==0)
			{
				# Some servers (like .org) don't send the nameservers in
				# the 'additional' field, so we gotta go get the damn
				# things manually.  Sigh.  Assholes.
				map
				{
					my $sname = $_;
					map { addcachedserver($domname, $sname, $_) }
							getinfo($sname, "A");
					map { addcachedserver($domname, $sname, $_) }
							getinfo($sname, "AAAA");
				} @srvrs;
			}
			else
			{
				foreach my $rr ($query->additional)
				{
					# NS records can only point to A/AAAA records
					next unless (($rr->type eq "A") || ($rr->type eq "AAAA"));
					next unless grep { $_ eq $rr->name } @srvrs;

					# Shove it into the list
					addcachedserver($domname, $rr->name, $rr->address);
				}
			}

			#print "${ind}- Got nameservers for '$domname'.\n";
		}

		# If we got our authoritative answer, we'll break out here.
		# Else, we'll loop back around with a new entry in our cached
		# servers array, and go from there.
	}

	return(%toret);
}


# Add a name to the global namecache
sub cachename
{
	my($name, $type, $info) = @_;

	# If it's not already in the list, stuff it in
	if(!(grep { $_ eq $info } @{$namecache{$name}->{$type}}))
	{ push(@{$namecache{$name}->{$type}}, $info); }
}


# Look up a specific record
sub getinfo
{
	my($name, $type) = @_;
	my $haveanswer=0;
	my @toret;
	my %toret_mx;
	my $toret_cname;

	# First, check the global cache
	# Simple!
	if( (defined($namecache{$name}))
			&& (defined($namecache{$name}->{$type})) )
	{
		return(@{$namecache{$name}->{$type}});
	}

	# Otherwise, we're gonna have to be more complex about it.  Sigh.
	while($haveanswer==0)
	{
		# Randomize to try and spread load
		$res->nameservers(randomize_list(getcachedservers($name)));

		my $query = $res->send("$name.", $type);
		if(!$query)
		{
			print "Query failed.  Status:\n";
			$res->print;
			print "\n";
			exit(1);
		}

		# First see what response we have
		my $header = $query->header;
		my $response = $header->rcode;
		if($response ne "NOERROR")
		{
			# Some kinda error.  Just bail.
			my $errhost = $query->answerfrom();
			$name =~ s/\.$//;
			print "Error looking up $type record for '$name' "
					. "from $errhost: $response\n";
			return;
		}

		# Check what we have
		# If this is authoritative, we're done.  Else, we're ready to
		# loop back around.
		if($header->aa==1)
		{
			# Yay, authoritative.
			# We're getting our answers.  Cool beans.
			# in the chain.  Cool beans.

			foreach my $rr ($query->answer)
			{
				if( ($rr->type eq "CNAME") && ($type ne "CNAME") )
				{
					# It's giving us a CNAME we didn't ask for.
					# We could probably handle this better.
					return;
				}

				next unless $rr->type eq $type;

				if( ($type eq "A") || ($type eq "AAAA") )
				{
					push(@toret, $rr->address);
					cachename($rr->name, $rr->type, $rr->address);
				}
				elsif($type eq "NS")
				{
					push(@toret, $rr->nsdname);
					cachename($rr->name, $rr->type, $rr->nsdname);
				}
				elsif($type eq "CNAME")
				{
					$toret_cname = $rr->cname;
					cachename($rr->name, $rr->type, $rr->cname);
				}
				# MX and SOA records don't get cached because I can't be
				# bothered to generalize enough.
				elsif($type eq "SOA")
				{ push(@toret, $rr->rname); }
				elsif($type eq "MX")
				{
					push(@{$toret_mx{$rr->preference}}, $rr->exchange);
				}
				# Forget the other types for now
			}

			# That should be about all we care about
			$haveanswer=1;
		}
		else
		{
			# Non-authoritative (but still non-error) answer.  This
			# should be a referral to an upstream DNS server.  So, just
			# check the 'authority' section.
			my $domname;
			my @srvrs;

			# This seems to happen in some cases...
			foreach my $rr ($query->answer)
			{
				if($rr->type eq "CNAME")
				{
					if($type ne "CNAME")
					{
						# It's giving us a CNAME we didn't ask for.
						# We could probably handle this better.
						return;
					}
				}
			}

			foreach my $rr ($query->authority)
			{
				next unless $rr->type eq "NS";
				$domname = $rr->name;
				push(@srvrs, $rr->nsdname);
			}

			if($header->arcount==0)
			{
				# Some servers (like .org) don't send the nameservers in
				# the 'additional' field, so we gotta go get the damn
				# things manually.  Sigh.  Assholes.
				map
				{
					my $sname = $_;
					map { addcachedserver($domname, $sname, $_) }
							getinfo($_, "A");
					map { addcachedserver($domname, $sname, $_) }
							getinfo($_, "AAAA");
				} @srvrs;
			}
			else
			{
				foreach my $rr ($query->additional)
				{
					# NS records can only point to A/AAAA records
					next unless (($rr->type eq "A") || ($rr->type eq "AAAA"));
					next unless grep { $_ eq $rr->name } @srvrs;

					# Shove it into the list
					addcachedserver($domname, $rr->name, $rr->address);
				}
			}

			#print "${ind}- Got nameservers for '$domname'.\n";
		}

		# If haveanswer==0, loop back around with (hopefully) another
		# cached level of nameserver IP's to work with.  Else, fall
		# through.
	}

	if($type eq "MX")
	{ return(%toret_mx); }
	elsif($type eq "CNAME")
	{ return($toret_cname); }
	else
	{ return(@toret); }
}


# Get closest DNS servers from cache
sub getcachedservers
{
	my($domain) = @_;
	my $nametrim;
	my @servers;

	if(!($domain=~/\.$/))
	{ $domain .= "."; }

	$nametrim = $domain;
	@servers = ();
	while(1)
	{
		#print "${ind}** Checking server cache for domain '$nametrim'... ";
		if(defined($srvcache{$nametrim}))
		{
			foreach my $hname (keys %{$srvcache{$nametrim}})
			{ push(@servers, @{$srvcache{$nametrim}->{$hname}}); }
			last;
		}
		#print "None.\n";
		(undef, $nametrim) = split(/\./, $nametrim, 2);
		if(length($nametrim)==0)
		{ $nametrim = "."; }
	}
	#print "Found!\n";

	return(@servers);
}


# Add a name to the DNS servercache
sub addcachedserver
{
	my($domain,$sname,$saddr) = @_;

	if(!($domain=~/\.$/))
	{ $domain .= "."; }

	# If it's not already in the list, stuff it in
	if(!(grep { $_ eq $sname } keys %{$srvcache{$domain}}))
	{ push(@{$srvcache{$domain}->{$sname}}, $saddr); }
}


# Map a set of addresses to a string
sub addrs2string
{
	my(@addrs) = @_;
	my $toret = "";

	map { $toret .= "$_ : " } @addrs;
	$toret =~ s/\ :\ $//;

	return($toret);
}

==> dinfo-1.1.tar.gz: dinfo-1.1/dinfo.pl <==
#!/usr/bin/perl -w
# Look up info for a domain
# $Fuller: myprograms/dinfo/dinfo.pl,v 1.9 2003/06/07 23:38:30 fullermd Exp $

use Net::DNS;
use strict;

# We'll use these a lot
my $i;
my %srvcache;
my %namecache;
my $ind = "  ";

# Create a resolver connection
my $res = Net::DNS::Resolver->new;
# Don't recurse
$res->recurse(0);

# First, some variables
# root server hints
my @roothints = (
		"128.9.0.107",
		"192.58.128.30",
		"193.0.14.129",
		"198.32.64.12",
		"202.12.27.33",
		"192.36.148.17",
		"192.203.230.10",
		"128.8.10.90",
		"198.41.0.4",
		"128.63.2.53",
		"192.33.4.12",
		"192.112.36.4",
		"192.5.5.241" );

# Look up the current root servers
$res->nameservers(randomize_list(@roothints));
my $query = $res->query(".", "NS");
if($query)
{
	foreach my $rr ($query->additional)
	{
		next unless (($rr->type eq "A")||($rr->type eq "AAAA"));
		addcachedserver(".", $rr->name, $rr->address);
		cachename($rr->name, $rr->type, $rr->address);
	}
} else {
	print "Error: Looking up root servers failed\n";
	exit(1);
}

# Loop over the domain we've requested
my $domain;
my @domain_servers;
$i=0;
while($domain = $ARGV[$i])
{
	# Get the info
	print "- Researching $domain...\n";

	# Full domain
	my %domitself = getdinfo("$domain");
	if(!%domitself)
	{ $i++; next; }

	# If it's a CNAME, chase it
	my @domcnameaddr;
	if(defined($domitself{"CNAME"}))
	{
		my $cnameto = $domitself{"CNAME"};
		@domcnameaddr = getinfo($cnameto, "A");
		if(!@domcnameaddr)
		{ @domcnameaddr = getinfo($cnameto, "AAAA"); }
	}

	# Look up its mailservers if we haven't already
	my %mxsrvs;
	if($domitself{"MX"})
	{
		foreach my $pref (keys %{$domitself{"MX"}})
		{
			map
			{
				my $srvname = $_;
				my @addrs;
				my $cnret = getinfo($srvname, "CNAME");
				if($cnret)
				{
					# It's a CNAME.  Chase it.
					push(@addrs, getinfo($cnret, "A"));
					push(@addrs, getinfo($cnret, "AAAA"));
					if(length(@addrs)>0)
					{
						$cnret .= " => " . addrs2string(@addrs);
					}
					$mxsrvs{$pref}->{$srvname} = $cnret;
				}
				else
				{
					push(@addrs, getinfo($srvname, "A"));
					push(@addrs, getinfo($srvname, "AAAA"));
					$mxsrvs{$pref}->{$srvname} = addrs2string(@addrs);
				}
			} @{$domitself{"MX"}->{$pref}};
		}
	}

	# www.$domain
	my @wwwinfo;
	my $wwwcname="";
	my @wwwcnameaddr;
	# Check for CNAME first
	$wwwcname = getinfo("www.$domain", "CNAME");
	if($wwwcname)
	{
		@wwwcnameaddr=getinfo($wwwcname, "A");
		if(!@wwwcnameaddr)
		{ @wwwcnameaddr=getinfo($wwwcname, "AAAA"); }
		if(!@wwwcnameaddr)
		{ @wwwcnameaddr = (); }
		@wwwinfo = ('foo');
	}
	else
	{
		push(@wwwinfo, getinfo("www.$domain", "A"));
		push(@wwwinfo, getinfo("www.$domain", "AAAA"));
	}

	# Show the info
	# Host A/AAAA/CNAME
	if( (defined($domitself{"A"})) || (defined($domitself{"AAAA"})) )
	{
		my $srvaddrs;
		my $srvaddrs6;
		$srvaddrs = addrs2string(@{$domitself{"A"}});
		$srvaddrs6 = addrs2string(@{$domitself{"AAAA"}});
		if($srvaddrs6)
		{
			if($srvaddrs)
			{ $srvaddrs .= " : "; }
			$srvaddrs .= $srvaddrs6;
		}
		print "${ind}o Host $domain has address ($srvaddrs)\n";
	}
	elsif(defined($domitself{"CNAME"}))
	{
		# CNAME'd instead
		my $cnameto;
		my $domcnameaddr="";
		$domcnameaddr .= addrs2string(@domcnameaddr);
		$cnameto = $domitself{"CNAME"};
		print "${ind}o Host $domain is a CNAME to $cnameto ($domcnameaddr)\n";
	}

	# Webserver
	if(@wwwinfo)
	{
		if($wwwcname)
		{
			my $wwwcnameaddr="";
			$wwwcnameaddr .= addrs2string(@wwwcnameaddr);
			print "${ind}o Host www.$domain is a CNAME to $wwwcname "
					. "($wwwcnameaddr)\n";
		}
		else
		{
			my $srvaddrs="";
			$srvaddrs .= addrs2string(@wwwinfo);
			print "${ind}o Host www.$domain has address ($srvaddrs)\n";
		}
	}

	# MX servers
	if(%mxsrvs)
	{
		print "${ind}o Mail exchangers:\n";
		foreach my $pref (sort { $a <=> $b } keys %mxsrvs)
		{
			my $firstone=1;
			foreach my $host (sort keys %{$mxsrvs{$pref}})
			{
				if($firstone)
				{ printf "${ind}${ind}* Pri=%3d ", $pref; $firstone=0; }
				else
				{ print  "${ind}${ind}          "; }
				printf "$host (%s)\n", $mxsrvs{$pref}->{$host};
			}
		}
	}

	# Nameservers (delegated)
	if($srvcache{"$domain."})
	{
		print "${ind}o Nameservers (from lower level nameservers):\n";
		foreach my $hname (sort keys %{$srvcache{"$domain."}})
		{
			my $nsaddrs;
			$nsaddrs = addrs2string(@{$srvcache{"$domain."}->{$hname}});
			print "${ind}${ind}* $hname ($nsaddrs)\n";
		}
	}

	# Nameservers (in zomefile)
	if($domitself{'NS'})
	{
		print "${ind}o Nameservers (in server zonefile):\n";
		map
		{
			my $nsaddrs;
			my $nsaddrs6;
			$nsaddrs=addrs2string(@{$domitself{'additional'}->{$_}->{"A"}});
			$nsaddrs6=addrs2string(@{$domitself{'additional'}->{$_}->{"AAAA"}});
			if($nsaddrs6)
			{
				if($nsaddrs)
				{ $nsaddrs .= " : "; }
				$nsaddrs .= $nsaddrs6;
			}
			print "${ind}${ind}* $_ ($nsaddrs)\n";
		} sort @{$domitself{'NS'}};
	}

	# SOA
	if($domitself{'contact'})
	{
		my $caddr = $domitself{'contact'};
		$caddr =~ s/\./@/;
		print "${ind}o SOA contact address: $caddr\n";
	}


	# Move on down the road.
	$i++;
}


# FUNCTIONS
# Randomize a list
sub randomize_list
{
	my(@thislist) = @_;
	my $nitems = @thislist;
	my $i;
	my $randret;
	my @toret;

	for($i=0 ; $i < $nitems ; $i++)
	{
		while($randret=rand($nitems))
		{
			if( (length($thislist[$randret]))>0 )
			{
				$toret[$i] = $thislist[$randret];
				$thislist[$randret]="";
				last;
			}
		}
	}

	return(@toret);
}


# Get all available info for the given domain.  We cache nameservers
# along the way, which is useful if we're looking up multiple domains.
# Be nice to the networks, people!
sub getdinfo
{
	my($dname) = @_;
	my $nametrim;
	my %toret;
	my $haveanswer=0;
	my $dummy;

	# Start moving up
	while($haveanswer==0)
	{
		# Randomize to try and spread load
		$res->nameservers(randomize_list(getcachedservers($dname)));

		my $query = $res->send("$dname.", "ANY");
		if(!$query)
		{
			print "Query failed.  Status:\n";
			$res->print;
			print "\n";
			return;
		}

		# First see what response we have
		my $header = $query->header;
		my $response = $header->rcode;
		if($response ne "NOERROR")
		{
			# Some kinda error.  Just bail.
			my $errhost = $query->answerfrom();
			print "Error looking up domain info for '$dname' "
					. "from $errhost: $response\n";
			return;
		}

		# Check what we have
		# If this is authoritative, we're done.  Else, we're ready to
		# loop back around.
		if($header->aa==1)
		{
			# Yay, authoritative.
			# We're getting our answers.  Cool beans.
			# in the chain.  Cool beans.

			foreach my $rr ($query->answer)
			{
				my $type = $rr->type;

				if( ($type eq "A") || ($type eq "AAAA") )
				{
					push(@{$toret{$type}}, $rr->address);
					cachename($rr->name, $rr->type, $rr->address);
				}
				elsif($type eq "NS")
				{
					push(@{$toret{$type}}, $rr->nsdname);
					cachename($rr->name, $rr->type, $rr->nsdname);
				}
				elsif($type eq "CNAME")
				{
					$toret{$type} = $rr->cname;
					cachename($rr->name, $rr->type, $rr->cname);
				}
				# MX and SOA records don't get cached because I can't be
				# bothered to generalize enough.
				elsif($type eq "SOA")
				{ $toret{'contact'} = $rr->rname; }
				elsif($type eq "MX")
				{ push(@{$toret{$type}->{$rr->preference}}, $rr->exchange); }
				# Forget the other types for now
			}

			# Get info from our "additional" section.
			foreach my $rr ($query->additional)
			{
				my $type = $rr->type;
				my $rname = $rr->name;

				if( ($type eq "A") || ($type eq "AAAA") )
				{
					push(@{$toret{'additional'}->{$rname}->{$type}},
							$rr->address);
				}
				# Forget the other types for now
			}

			# That should be about all we care about
			$haveanswer=1;
		}
		else
		{
			# Non-authoritative (but still non-error) answer.  This
			# should be a referral to an upstream DNS server.  So, just
			# check the 'authority' section.
			my $domname;
			my @srvrs;

			foreach my $rr ($query->authority)
			{
				next unless $rr->type eq "NS";
				$domname = $rr->name;
				push(@srvrs, $rr->nsdname);
			}

			if($header->arcount==0)
			{
				# Some servers (like .org) don't send the nameservers in
				# the 'additional' field, so we gotta go get the damn
				# things manually.  Sigh.  Assholes.
				map
				{
					my $sname = $_;
					map { addcachedserver($domname, $sname, $_) }
							getinfo($sname, "A");
					map { addcachedserver($domname, $sname, $_) }
							getinfo($sname, "AAAA");
				} @srvrs;
			}
			else
			{
				foreach my $rr ($query->additional)
				{
					# NS records can only point to A/AAAA records
					next unless (($rr->type eq "A") || ($rr->type eq "AAAA"));
					next unless grep { $_ eq $rr->name } @srvrs;

					# Shove it into the list
					addcachedserver($domname, $rr->name, $rr->address);
				}
			}

			#print "${ind}- Got nameservers for '$domname'.\n";
		}

		# If we got our authoritative answer, we'll break out here.
		# Else, we'll loop back around with a new entry in our cached
		# servers array, and go from there.
	}

	return(%toret);
}


# Add a name to the global namecache
sub cachename
{
	my($name, $type, $info) = @_;

	# If it's not already in the list, stuff it in
	if(!(grep { $_ eq $info } @{$namecache{$name}->{$type}}))
	{ push(@{$namecache{$name}->{$type}}, $info); }
}


# Look up a specific record
sub getinfo
{
	my($name, $type) = @_;
	my $haveanswer=0;
	my @toret;
	my %toret_mx;
	my $toret_cname;

	# First, check the global cache
	# Simple!
	if( (defined($namecache{$name}))
			&& (defined($namecache{$name}->{$type})) )
	{
		return(@{$namecache{$name}->{$type}});
	}

	# Otherwise, we're gonna have to be more complex about it.  Sigh.
	while($haveanswer==0)
	{
		# Randomize to try and spread load
		$res->nameservers(randomize_list(getcachedservers($name)));

		my $query = $res->send("$name.", $type);
		if(!$query)
		{
			print "Query failed.  Status:\n";
			$res->print;
			print "\n";
			exit(1);
		}

		# First see what response we have
		my $header = $query->header;
		my $response = $header->rcode;
		if($response ne "NOERROR")
		{
			# Some kinda error.  Just bail.
			my $errhost = $query->answerfrom();
			$name =~ s/\.$//;
			print "Error looking up $type record for '$name' "
					. "from $errhost: $response\n";
			return;
		}

		# Check what we have
		# If this is authoritative, we're done.  Else, we're ready to
		# loop back around.
		if($header->aa==1)
		{
			# Yay, authoritative.
			# We're getting our answers.  Cool beans.
			# in the chain.  Cool beans.

			foreach my $rr ($query->answer)
			{
				if( ($rr->type eq "CNAME") && ($type ne "CNAME") )
				{
					# It's giving us a CNAME we didn't ask for.
					# We could probably handle this better.
					return;
				}

				next unless $rr->type eq $type;

				if( ($type eq "A") || ($type eq "AAAA") )
				{
					push(@toret, $rr->address);
					cachename($rr->name, $rr->type, $rr->address);
				}
				elsif($type eq "NS")
				{
					push(@toret, $rr->nsdname);
					cachename($rr->name, $rr->type, $rr->nsdname);
				}
				elsif($type eq "CNAME")
				{
					$toret_cname = $rr->cname;
					cachename($rr->name, $rr->type, $rr->cname);
				}
				# MX and SOA records don't get cached because I can't be
				# bothered to generalize enough.
				elsif($type eq "SOA")
				{ push(@toret, $rr->rname); }
				elsif($type eq "MX")
				{
					push(@{$toret_mx{$rr->preference}}, $rr->exchange);
				}
				# Forget the other types for now
			}

			# That should be about all we care about
			$haveanswer=1;
		}
		else
		{
			# Non-authoritative (but still non-error) answer.  This
			# should be a referral to an upstream DNS server.  So, just
			# check the 'authority' section.
			my $domname;
			my @srvrs;

			# This seems to happen in some cases...
			foreach my $rr ($query->answer)
			{
				if($rr->type eq "CNAME")
				{
					if($type ne "CNAME")
					{
						# It's giving us a CNAME we didn't ask for.
						# We could probably handle this better.
						return;
					}
				}
			}

			foreach my $rr ($query->authority)
			{
				next unless $rr->type eq "NS";
				$domname = $rr->name;
				push(@srvrs, $rr->nsdname);
			}

			if($header->arcount==0)
			{
				# Some servers (like .org) don't send the nameservers in
				# the 'additional' field, so we gotta go get the damn
				# things manually.  Sigh.  Assholes.
				map
				{
					my $sname = $_;
					map { addcachedserver($domname, $sname, $_) }
							getinfo($_, "A");
					map { addcachedserver($domname, $sname, $_) }
							getinfo($_, "AAAA");
				} @srvrs;
			}
			else
			{
				foreach my $rr ($query->additional)
				{
					# NS records can only point to A/AAAA records
					next unless (($rr->type eq "A") || ($rr->type eq "AAAA"));
					next unless grep { $_ eq $rr->name } @srvrs;

					# Shove it into the list
					addcachedserver($domname, $rr->name, $rr->address);
				}
			}

			#print "${ind}- Got nameservers for '$domname'.\n";
		}

		# If haveanswer==0, loop back around with (hopefully) another
		# cached level of nameserver IP's to work with.  Else, fall
		# through.
	}

	if($type eq "MX")
	{ return(%toret_mx); }
	elsif($type eq "CNAME")
	{ return($toret_cname); }
	else
	{ return(@toret); }
}


# Get closest DNS servers from cache
sub getcachedservers
{
	my($domain) = @_;
	my $nametrim;
	my @servers;

	if(!($domain=~/\.$/))
	{ $domain .= "."; }

	$nametrim = $domain;
	@servers = ();
	while(1)
	{
		#print "${ind}** Checking server cache for domain '$nametrim'... ";
		if(defined($srvcache{$nametrim}))
		{
			foreach my $hname (keys %{$srvcache{$nametrim}})
			{ push(@servers, @{$srvcache{$nametrim}->{$hname}}); }
			last;
		}
		#print "None.\n";
		(undef, $nametrim) = split(/\./, $nametrim, 2);
		if(length($nametrim)==0)
		{ $nametrim = "."; }
	}
	#print "Found!\n";

	return(@servers);
}


# Add a name to the DNS servercache
sub addcachedserver
{
	my($domain,$sname,$saddr) = @_;

	if(!($domain=~/\.$/))
	{ $domain .= "."; }

	# If it's not already in the list, stuff it in
	if(!(grep { $_ eq $sname } keys %{$srvcache{$domain}}))
	{ push(@{$srvcache{$domain}->{$sname}}, $saddr); }
}


# Map a set of addresses to a string
sub addrs2string
{
	my(@addrs) = @_;
	my $toret = "";

	map { $toret .= "$_ : " } @addrs;
	$toret =~ s/\ :\ $//;

	return($toret);
}
